=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed Dec. 2008 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:  Lib6Mark.rb
# Original Date	:  23 Dec 2008 - version 3.1
# Type			:  Script library part of the LibFredo6 shared libraries
# Description	:  Contains some utilities for drawing standard predefined marks and shapes
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module G6

#--------------------------------------------------------------------------------------------------------------
# Class DrawMark: Predefined marks for 2D and 3D drawing 
#--------------------------------------------------------------------------------------------------------------			 

class AllDrawMark

Traductor_DrawMark_Elt = Struct.new "Traductor_DrawMark_Elt", :lpt, :color, :width, :stipple, :gl_type

def creation_mark
	@list_elts = []
	@stipple = ""
	@color = 'black'
	@width = 1
end

def draw_at_point2d(view, pt2d, color=nil)
	size = (@size) ? @size : 1
	t = Geom::Transformation.translation(pt2d) * Geom::Transformation.scaling(size)

	@list_elts.each do |elt|
		lpt = elt.lpt
		lpt2d = lpt.collect { |pt| t * pt }
		view.drawing_color = (color) ? color : ((elt.color) ? elt.color : @color)
		view.line_width = (elt.width) ? elt.width : @width
		view.line_stipple = (elt.stipple) ? elt.stipple : @stipple
		view.draw2d elt.gl_type, lpt2d
	end	
end

def draw_at_xy(view, x, y)
	draw_at_point2d view, Geom::Point3d.new(x, y, 0)
end

def draw_at_point3d(view, pt3d, color=nil)
	draw_at_point2d view, view.screen_coords(pt3d), color
end

def create_element(lpt, gl_type, color=nil, width=nil, stipple=nil)
	elt = Traductor_DrawMark_Elt.new
	elt.lpt = lpt
	elt.gl_type = gl_type
	elt.color = color
	elt.width = width
	elt.stipple = stipple
	elt
end

def create_line(lpt)
	create_element lpt, GL_LINE_STRIP
end

end	#class AllDrawMark

#--------------------------------------------------------------------------------------------------------------
# Class DrawMark_Forbidden: Forbidden sign 
#--------------------------------------------------------------------------------------------------------------			 

class DrawMark_Forbidden < AllDrawMark

def initialize(size=nil, color=nil, width=nil)
	creation_mark
	@size = (size) ? size : 5
	@color = (color) ? color : 'red'
	@width = (width) ? width : 2
	
	#Main circle
	n = 12
	pi = Math::PI
	radius = 10
	circle = []
	for i in 0..n
		angle = 2 * pi * i / n
		cosinus = Math::cos angle
		sinus = Math::sin angle
		circle.push Geom::Point3d.new(cosinus, sinus, 0)
	end
	bar = [circle[5], circle[11]]
	@list_elts.push create_line(circle)
	@list_elts.push create_line(bar)
end

end	#class DrawMark_Forbidden

#--------------------------------------------------------------------------------------------------------------
# Class DrawMark_FourArrows: 4 arrows 
#--------------------------------------------------------------------------------------------------------------			 

class DrawMark_FourArrows < AllDrawMark

def initialize(size=nil, color=nil, width=nil)
	creation_mark
	@size = (size) ? size : 7
	@color = (color) ? color : 'blue'
	@width = (width) ? width : 2
	
	#Shape
	dec = 0.4
	pt1 = Geom::Point3d.new -1, 0, 0
	pt2 = Geom::Point3d.new 1, 0, 0
	pt3 = Geom::Point3d.new 0, -1, 0
	pt4 = Geom::Point3d.new 0, 1, 0
	@list_elts.push create_line([pt1, pt2])
	@list_elts.push create_line([pt3, pt4])
	@list_elts += make_arrow(pt1, -dec, 'X')
	@list_elts += make_arrow(pt2, dec, 'X')
	@list_elts += make_arrow(pt3, -dec, 'Y')
	@list_elts += make_arrow(pt4, dec, 'Y')
end

def make_arrow(pt2d, dec, code)
	lselt = []
	case code
	when 'X'
		x1 = pt2d.x - dec
		y1 = pt2d.y + dec
		y2 = pt2d.y - dec
		lselt.push create_line([pt2d, Geom::Point3d.new(x1, y1, 0)])
		lselt.push create_line([pt2d, Geom::Point3d.new(x1, y2, 0)])
	when 'Y'
		y1 = pt2d.y - dec
		x1 = pt2d.x + dec
		x2 = pt2d.x - dec
		lselt.push create_line([pt2d, Geom::Point3d.new(x1, y1, 0)])
		lselt.push create_line([pt2d, Geom::Point3d.new(x2, y1, 0)])
	end
	lselt	
end

end	#class DrawMark_FourArrows

#--------------------------------------------------------------------------------------------------------------
# Class DrawMark_H2Arrows: 2 arrows horizontal
#--------------------------------------------------------------------------------------------------------------			 

class DrawMark_H2Arrows < G6::AllDrawMark

def initialize(size=nil, color=nil, width=nil)
	creation_mark
	@size = (size) ? size : 7
	@color = (color) ? color : 'darkgreen'
	@width = (width) ? width : 2
	
	#Shape
	dec = 0.4
	pt1 = Geom::Point3d.new -1, 0, 0
	pt2 = Geom::Point3d.new 1, 0, 0
	@list_elts.push create_line([pt1, pt2])
	@list_elts += make_arrow(pt1, -dec)
	@list_elts += make_arrow(pt2, dec)
end

def make_arrow(pt2d, dec)
	lselt = []
	x1 = pt2d.x - dec
	y1 = pt2d.y + dec
	y2 = pt2d.y - dec
	lselt.push create_line([pt2d, Geom::Point3d.new(x1, y1, 0)])
	lselt.push create_line([pt2d, Geom::Point3d.new(x1, y2, 0)])
	lselt	
end

end	#class DrawMark_FourArrows

#--------------------------------------------------------------------------------------------------------------
# Class DrawMark_Curve: Half circle
#--------------------------------------------------------------------------------------------------------------			 

class DrawMark_Curve < G6::AllDrawMark

def initialize(size=nil, color=nil, width=nil)
	creation_mark
	@size = (size) ? size : 9
	@color = (color) ? color : 'blue'
	@width = (width) ? width : 2
	
	#Shape
	lpt = []
	n = 8
	angle = Math::PI / n
	for i in 0..n
		lpt.push Geom::Point3d.new(Math.sin(angle * i), Math.cos(angle * i), 0)
	end	
	for i in 0..n-1
		@list_elts.push create_line([lpt[i], lpt[i+1]])
	end	
end

end	#class DrawMark_Curve

#--------------------------------------------------------------------------------------------------------------
# Class ProtractorShape: Input tool to show angle selection
#--------------------------------------------------------------------------------------------------------------			 				   
		
class ShapeProtractor

attr_reader :origin, :normal, :base_angle, :basedir, :cur_angle, :curdir, :radius_bigcircle

def initialize(*hargs)
	@origin = nil
	@cur_angle = 0
	@base_angle = 0
	@normal = nil
	@curdir = nil
	@basedir = nil
	@default_color = 'black'

	#Parsing the arguments
	hargs.each { |harg| harg.each { |key, value|  parse_args(key, value) } if harg.class == Hash }
	
	#geometry of the protractor
	@bigcircle = []
	@graduations = []
	@inner_up = []
	@inner_down = []
	n = 24
	pi = Math::PI
	radius = @radius2d = 200.cm
	fact_grad = 0.9
	fact_grad2 = 0.80
	fact_inner = 0.75
	limit = Math::sin(pi / 12)
	
	#defining the Protractor
	for i in 0..n
		angle = 2 * pi * i / n
		cosinus = Math::cos angle
		sinus = Math::sin angle
		pt_ext = Geom::Point3d.new radius * cosinus, radius * sinus, 0
		@bigcircle.push pt_ext
		fac = (i == 6 || i == 18) ? fact_grad2 : fact_grad
		pt_int = Geom::Point3d.new radius * fac * cosinus, radius * fac * sinus, 0
		@graduations.push pt_int
		@graduations.push pt_ext
		if (sinus.abs >= limit)
			pt_ext = Geom::Point3d.new radius * fact_inner * cosinus, radius * fact_inner * sinus, 0
			if (sinus > 0.0)
				@inner_up.push pt_ext
			else
				@inner_down.push pt_ext
			end	
		end
	end
	@inner_up.push @inner_up[0]
	@inner_down.push @inner_down[0]
	
	dec = radius * 0.1
	@inner_cross = [Geom::Point3d.new(dec, 0, 0), Geom::Point3d.new(-dec, 0, 0), Geom::Point3d.new(0, dec, 0), Geom::Point3d.new(0, -dec, 0)]
end

#Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /default_color/i
		@default_color = value
	end	
end

#Set the face under the protractor (if any)
def set_face(face)
	@face = face
	@color = Traductor::Couleur.color_vector @normal, @default_color, @face
	face
end

#specify the placement of the Protactor
def set_placement(origin, normal, basedir=nil, face=nil)
	@origin = origin
	view = Sketchup.active_model.active_view
	@normal = (normal % view.camera.direction < 0) ? normal : normal.reverse
	@axe0 = @normal.axes[0] 
	@axe1 = @normal.axes[1] 
	@face = face
	set_basedir basedir
	@color = Traductor::Couleur.color_vector @normal, @default_color, @face
end

#Set the base direction for the protractor, and thus allow to rotate it
def set_basedir(basedir, ptinf=nil)
	@basedir = (basedir && basedir.valid?) ? (@normal * (basedir * @normal)) : @axe0
	check_inference @basedir, 'B', ptinf if ptinf
	recompute_after_base_changed
end

#Set the base direction via an angle in radians (absolute or increment)
def set_baseangle(angle, incr=false)
	angle += @base_angle if incr
	@basedir = Geom::Transformation.rotation(@origin, @normal, angle) * @axe0
	recompute_after_base_changed
end

#Recompute the parameters after the base angle was changed
def recompute_after_base_changed
	#computing the base angle
	angle = @axe0.angle_between(@basedir)
	angle = -angle if @basedir % @axe1 < 0
	@base_angle = angle
	
	#Computing the current angle
	@otherdir = @normal * @basedir
	@curdir = Geom::Transformation.rotation(@origin, @normal, @cur_angle) * @basedir
	@basedir
end

#Set the current direction for the protractor
def set_curdir(curdir, ptinf=false)
	@curdir = (curdir) ? (@normal * (curdir * @normal)) : @basedir
	check_inference @curdir, 'C', ptinf if ptinf
	recompute_after_cur_changed
end

#Set the current direction via an angle in radians (absolute or increment)
def set_curangle(angle, incr=false)
	angle += @cur_angle if incr
	@curdir = Geom::Transformation.rotation(@origin, @normal, angle) * @basedir
	recompute_after_cur_changed
end

#Recompute the parameters after the current angle was changed
def recompute_after_cur_changed
	#computing the base angle
	angle = @basedir.angle_between(@curdir)
	angle = -angle if @curdir % @otherdir < 0
	@cur_angle = angle
	@curdir
end

#Check inference for a given vector (direction and length)
def check_inference(vecdir, inftype, ptinf)
	return vecdir unless ptinf && vecdir && vecdir.valid? && @radius_bigcircle && @normal.valid?
	pt = ptinf.project_to_plane [@origin, @normal]
	factor = @origin.distance(pt) / @radius_bigcircle
	if inftype == 'B'
		angle = @axe0.angle_between vecdir
		angle = -angle if vecdir % @axe1 < 0
		angle = adjust_angle angle, factor
		@basedir = Geom::Transformation.rotation(@origin, @normal, angle) * @axe0
	else	
		angle = @basedir.angle_between vecdir
		angle = -angle if vecdir % @otherdir < 0
		angle = adjust_angle angle, factor
		@curdir = Geom::Transformation.rotation(@origin, @normal, angle) * @basedir
	end
end

def get_radius2d
	@radius2d
end

#Adjust angleto round value,depending on distance of input point to Origin
def adjust_angle(angle, factor)
	imax = 3
	fac = factor.round + 1
	incr = (fac <= 1.5) ? 15.0 : ((fac <= imax ) ? 5.0 : 1.0)
	tolerance = (fac <= imax) ? incr : (2 * incr / fac)
	dangle = angle.radians
	dangle = dangle.round if fac < imax
	a = dangle / incr
	around = a.round * incr
	afinal = ((dangle - around).abs < tolerance) ? around : dangle
	afinal.degrees
end

#Draw method for tool
def draw(view, factor=1.0)
	return unless @origin
	
	#Compute the right scale to keep the protractor the same size
	size = view.pixels_to_model 1, @origin
	size *= factor if factor
	tdir = Geom::Transformation.axes @origin, @basedir, @otherdir, @normal
	return unless tdir
	t = tdir * Geom::Transformation.scaling(size)
	
	#draw the protractor
	view.drawing_color = @color 
	view.line_stipple = ""
	view.line_width = 1

	pts = @bigcircle.collect { |pt| t * pt }
	@radius_bigcircle = @origin.distance pts[0]
	draw2d view, pts, GL_LINE_STRIP		#bug in Sketchup, need a first dummy drawing in 2d
	draw2d view, pts, GL_LINE_STRIP
	
	pts = []
	pts = @inner_up.collect { |pt| t * pt }
	draw2d view, pts, GL_LINE_STRIP
	
	pts = @inner_down.collect { |pt| t * pt }
	draw2d view, pts, GL_LINE_STRIP
	
	pts = @graduations.collect { |pt| t * pt }
	draw2d view, pts, GL_LINES
	
	pts = @inner_cross.collect { |pt| t * pt }
	draw2d view, pts, GL_LINES
end

def draw2d(view, pts, type)
	pts2d = pts.collect { |pt| p = view.screen_coords(pt) ; p.z = 0 ; p }
	view.draw2d type, pts2d
end

#Draw the dashed line indicating the current angle, with clipping on the view
def draw_dashed_line(view, color=nil, stipple=nil, width=nil)
	return unless @curdir.valid?
	view.drawing_color = Traductor::Couleur.color_vector @curdir, @default_color, @face
	view.line_stipple = (stipple) ? stipple : "_"
	view.line_width = (width) ? width : 1
	pt1 = view.screen_coords @origin
	pt2 = view.screen_coords @origin.offset(@curdir, 1)
	view.draw2d GL_LINES, clip_to_box(view, pt1, pt2)
end

def clip_to_box(view, pt1, pt2)	
	x1 = pt1.x
	y1 = pt1.y
	x2 = pt2.x
	y2 = pt2.y
	vpwidth = view.vpwidth
	vpheight = view.vpheight
	ptbox = [Geom::Point3d.new(0, 0, 0), Geom::Point3d.new(0, 0, 0)]	

	ay = (y1 - y2)
	ax = (x1 - x2)
	
	if (ay == 0)
		ptbox[0].x = 0
		ptbox[0].y = ptbox[1].y = y1
		ptbox[1].x = vpwidth
	elsif (ax == 0)
		ptbox[0].x = ptbox[1].x = x1
		ptbox[0].y = 0
		ptbox[1].y = vpheight
	else
		a = ay / ax
		i = 0
		v = y1 - a * x1						#left edge
		if (v >= 0 && v <= vpheight)		
			ptbox[i].x = 0
			ptbox[i].y = v
			i += 1	
		end	
		v = y1 + a * (vpwidth - x1)		#right edge
		if (v >= 0 && v <= vpheight)
			ptbox[i].x = vpwidth
			ptbox[i].y = v
			return ptbox if ((i += 1) > 1)
		end
		v = x1 - y1 / a 					#top edge
		if (v >= 0 && v <= vpwidth)
			ptbox[i].x = v
			ptbox[i].y = 0
			return ptbox if ((i += 1) > 1)
		end	
		v = x1 + (vpheight - y1) / a		#bottom edge
		if (v >= 0 && v <= vpwidth)
			ptbox[i].x = v
			ptbox[i].y = vpheight
		end
	end
	ptbox
end

end	#class ProtractorShape

#--------------------------------------------------------------------------------------------------------------
# Class ShapeGrid: Input tool to show angle selection
#--------------------------------------------------------------------------------------------------------------			 				   
		
class ShapeGrid

def initialize(hnx=2, hny=nil)
	hny = hnx unless hny
	@grid_lines = []
	@grid_contour = []
	for i in -hnx..hnx
		@grid_lines.push Geom::Point3d.new(i, -hny, 0), Geom::Point3d.new(i, hny, 0)
	end	
	for i in -hny..hny
		@grid_lines.push Geom::Point3d.new(-hnx, i, 0), Geom::Point3d.new(hnx, i, 0)
	end
	@grid_contour = [Geom::Point3d.new(-hnx, -hny, 0), Geom::Point3d.new(hnx, -hny, 0),
	                 Geom::Point3d.new(hnx, hny, 0), Geom::Point3d.new(-hnx, hny, 0)]
	@line_width = 1
	@line_stipple = ''
	@poly_frame_width = 2
	@poly_frame_stipple = ''
	
	color = Sketchup::Color.new 'yellow'
	color.alpha = 0.5
	@poly_back_color = color
	@poly_frame_color = 'blue'
	@line_color = 'black'

end

def set_line_param(line_color=nil, line_width=nil, line_stipple=nil)
	@line_color = line_color if line_color
	@line_width = line_width if line_width
	@line_stipple = line_stipple if line_stipple	
end

def set_polygon_param(bk_color=nil, fr_color=nil, fr_width=nil, fr_stipple=nil)
	@poly_back_color = bk_color if bk_color
	@poly_frame_color = fr_color if fr_color
	@poly_frame_width = fr_width if fr_width
	@poly_frame_stipple = fr_stipple if fr_stipple
end

#Drawing methods, 3D and 2D
def draw_generic(in_3d, view, origin, normal, xdir=nil, pix_size=15)
	ydir = (xdir) ? normal * xdir : nil
	if ydir && ydir.valid?
		axes = [xdir, ydir, normal]
	else
		axes = normal.axes
	end	
	tr_axes = Geom::Transformation.axes origin, *axes
	size = view.pixels_to_model pix_size, origin
	ts = Geom::Transformation.scaling size
	t = tr_axes * ts
	
	if @poly_back_color
		view.drawing_color = @poly_back_color
		contour = @grid_contour.collect { |pt| t * pt }
		if in_3d
			view.draw GL_POLYGON, contour.collect { |pt| G6.small_offset view, pt }
		else	
			view.draw2d GL_POLYGON, contour.collect { |pt| view.screen_coords(pt) }
		end	
	end
	
	if @line_color
		view.drawing_color = @line_color
		view.line_width = @line_width
		view.line_stipple = @line_stipple
		lines2d = @grid_lines.collect { |pt| t * pt }
		if in_3d
			view.draw GL_LINES, lines2d.collect { |pt| G6.small_offset view, pt }
		else	
			view.draw2d GL_LINES, lines2d.collect { |pt| view.screen_coords(pt) }
		end	
	end	
	
	if @poly_frame_color
		view.drawing_color = @poly_frame_color
		view.line_width = @poly_frame_width
		view.line_stipple = @poly_frame_stipple
		contour = @grid_contour.collect { |pt| t * pt }
		if in_3d
			view.draw GL_LINE_LOOP, contour.collect { |pt| G6.small_offset view, pt, 2 }
		else	
			view.draw2d GL_LINE_LOOP, contour.collect { |pt| view.screen_coords(pt) }
		end	
	end
	
end

def draw2d(view, origin, normal, xdir, pix_size)
	draw_generic false, view, origin, normal, xdir, pix_size
end	

def draw(view, origin, normal, xdir, pix_size)
	draw_generic true, view, origin, normal, xdir, pix_size
end	

end	#class ShapeGrid

end #Module G6

