=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Copyright 2011 Fredo6 - Designed and written June 2011 by Fredo6
#
# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6VCB.rb
# Original Date	:   13 June 2011
# Description	:   Manage VCB inputs
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# VCB Input management
#--------------------------------------------------------------------------------------------------------------			 				   
#--------------------------------------------------------------------------------------------------------------	
  
class VCB

InputFormat = Struct.new :symb, :letter, :typenum, :suffixes, :pattern, :validate_proc, :description

#--------------------------------------------------------------------------------------------------------------
# Initialization
#--------------------------------------------------------------------------------------------------------------			 				   

#Initialization	
def initialize__
	@lst_inputs = []
end

#Declare an input format. Conventions for specs is are <typenum>_<letter>_<suffixes> (or a list of specs)
def declare_input_format(symb, format_specs, description=nil, &validate_proc)
	format_specs = [format_specs] unless format_specs.class == Array
	format_specs.each do |specs|
		declare_single_format symb, specs, description, validate_proc
	end
end

#Declare an input format. Conventions are <typenum>_<letter>_<suffixes>
def declare_single_format(symb, specs, description, validate_proc)
	#Getting the specs
	typenum, letter, suffixes = specs.strip.split '_'
	typenum = (typenum && typenum != '') ? typenum.downcase.intern : nil
	
	#Letter, with special cases
	lpat = ""
	if letter && letter != ""
		letter.split('').each do |c|
			case c
			when '/', '*', '?', '+', '-'
				lpat += "\\" + c
			else
				lpat += "#{c.downcase}"
			end	
		end
	end
	
	#Creating the regular pattern
	numpat = ".*[0-9)]"
	totalpat = "\\A"	
	if typenum == :l
		totalpat += "(.*[0-9]+'*\"*[a-z]*)"
	elsif typenum == :slash
		totalpat += "\\/(-*[0-9]+)"
	elsif typenum == :star
		totalpat += "\\*(-*[0-9]+)"
	elsif typenum == :a
		totalpat += "(#{numpat}[dgr\\%])"
	elsif typenum == :aa
		totalpat += "(#{numpat}(dd|gg|rr|\\%\\%))"
	elsif typenum
		totalpat += "(#{numpat})#{lpat}"
	else	
		totalpat += "()#{lpat}"
	end			
	sfpat = (suffixes) ? "([#{suffixes.downcase}]?)" : ""	
	totalpat += sfpat + "\\Z"
	
	#Creating the structure
	input = InputFormat.new
	input.symb = symb
	input.validate_proc = validate_proc
	input.description = description
	input.typenum = typenum
	input.letter = letter
	input.suffixes = suffixes
	input.pattern = Regexp.new totalpat, Regexp::IGNORECASE
	
	if typenum.to_s.length > 2
		@lst_inputs.unshift input
	else
		@lst_inputs.push input
	end
	
	input
end

#Go through the results. Each items has 3 parameters
#   - <symb>: the symbol of the variable
#   - <val>: the value
#   - <suffix>: the optional suffix, if any
def each_result
	@lst_results.each { |a| yield *a }
end

#Go through the errors. Each items has 2 parameters
#   - <symb>: the symbol of the variable or nil if the error is in the parsing
#   - <s>: the faulty string item
def each_error
	@lst_errors.each { |a| yield *a }
end

#Process the parsing of the VCB text
#Return the number of errors
def process_parsing(text)
	@lst_results = []
	@lst_errors = []
	text = text.strip + ' '
	return unless text
	text = text.gsub(/'\s+[0-9]+\.*[0-9]*"/) { |a| a.sub(/\s+/, '___') }
	text = text.gsub(/'\s*[0-9]*\s+[0-9]+\/[0-9]+/) { |a| a.gsub(/\s+/, '___') }
	litems = text.split(/\s*;\s*;*|\s+;*/)	
	litems.each { |s| check_item s.strip.gsub(/___/, ' ') }
	@lst_errors.length
end
 
#Check an item of the VCB string 
def check_item(s)
	lerrors = []
	@lst_inputs.each do |input|
		if s.strip == ''
			next if input.typenum != :l
			@lst_results.push [input.symb, 0, nil]
			return
		end
		symb = input.symb
		if s =~ input.pattern
			suffix = $2
			if $1 == ""
				val = nil
			else	
				val = validation_value(symb, input.typenum, $1, input.validate_proc)
				if !val
					lerrors.push [symb, s]
					next
				end
			end	
			@lst_results.push [symb, val, suffix]
			return
		end
	end	
	if lerrors.empty?
		@lst_errors.push [nil, s]
	else
		@lst_errors += lerrors
	end	
end

#Validate individual values and convert them to their right type
def validation_value(symb, typenum, val, validate_proc)
	case typenum
	when :l
		val = Traductor.string_to_length_formula val
	when :f
		val = Traductor.string_to_float_formula val
	when :i, :star, :slash
		val = Traductor.string_to_integer_formula val
	when :aa
		val = Traductor.string_to_angle_degree_formula val[0..-2]
	when :a
		val = Traductor.string_to_angle_degree_formula val
	end		
	val = validate_proc.call(val) if val && validate_proc
	val
end
 
end	#class VCB

#--------------------------------------------------------------------------------------------------------------
#--------------------------------------------------------------------------------------------------------------
# VCBDialog: dialog box to replace VCB
#--------------------------------------------------------------------------------------------------------------			 				   
#--------------------------------------------------------------------------------------------------------------	

class VCBDialog

#Calculate the dialog box
def initialize__(unique_key, ini_value, call_back_proc)
	#Initialization
	@call_back_proc = call_back_proc
	@vcb_final = nil
	@title = "VCB"

	#Computing the HTML
	@id_input = 'INPUT'
	@vcb_value = ini_value
	html = format_html
	
	#Creating the dialog box
	@wdlg_key = 'VCBDialog'
	@wdlg = Traductor::Wdlg.new @title, @wdlg_key, false
	@wdlg.set_unique_key unique_key
	@wdlg.set_html html
	@wdlg.set_size 500, 100
	@wdlg.set_background_color 'AliceBlue'
	@wdlg.set_callback self.method('callback_dialog_top') 
	@wdlg.set_on_close { on_close } 
	@wdlg.initial_focus(@id_input)
	@wdlg.show_modal
end

def refresh_window
	@wdlg.set_html format_html_top if @wdlg
end

#Build the HTML for TOP Dialog
def format_html
	
	#Creating the HTML stream	
	html = Traductor::HTML.new
		
	#style used in the dialog box
	html.create_style 'Button', nil, 'K: black', 'F-SZ: 10'
	html.create_style 'Label', nil, 'B', 'K: black', 'F-SZ: 11'
	html.create_style 'Input', nil, 'K: navy', 'F-SZ: 11'
	
	#Creating the main table	
	tip = "VCB input"
	txlabel = "Entry"
	label = HTML.format_span(txlabel, "", "Label", nil, tip)
	but_ok = HTML.format_button T6[:T_BUTTON_Done], id="ButtonDone", 'Button', nil
	field = HTML.format_input(@vcb_value, "40", @id_input, 'Input', nil, tip)
	
	html.body_add "<table width='99%' cellpadding='1px'>"
	html.body_add "<tr><td>#{label}</td><td>#{field}</td><td>#{but_ok}</td></tr></table>"
	
	#Returning the HTML object
	html	
end

#Call back for Top Dialog
def callback_dialog_top(event, type, id, svalue)
	case event
	
	#Initialistaion
	when /wonload/i
		@@top_dialog = self
		@wdlg.execute_script "j6_cursor_at_end ('#{@id_input}')"
	
	#Command buttons
	when /onchange/i
		case id
		when @id_input
			puts "Input = #{svalue}"
			@vcb_value = svalue
		end
		
	when /onclick/i
		case id
		when 'ButtonDone'
			puts "Done value = #{@vcb_value}"
			@vcb_final = @vcb_value
			@wdlg.close
		end
		
	when /onKeyUp/i
		lskey = svalue.split '*'
		case lskey[0].to_i
		when 13
			puts "press return"
			@vcb_final = @vcb_value
		when 27
			puts "press cancel"
			@vcb_value = nil
		else
			return nil
		end
		@wdlg.close	
		
	end
end

#Force a close of the dialog boxes
def close_dialog
	@wdlg.close
end
	
#Notification when the top dialog box closes	
def on_close()
	puts "on close"
	@@top_dialog = nil
	@call_back_proc.call(@vcb_final) if @vcb_final && @call_back_proc
end

end	#class VCBDialog

end	#End Module Traductor
